char *visnum();
char *vispnum();
char *vispan();
char *vispoly();
char *visvelosw();
char *vismode();
char *vistwofour();
char *vismute();
char *vispolym();
char *visvbwh();
char *viscrv();
char *viskeyshift();
char *visfinetune();
char *viscpitch();
char *vislpitch();
char *visonoff();
char *visonoffk1();
char *visonoffk2();
char *visonoffk3();
char *visonoffk4();
char *viscsfnk1();
char *viscsfnk2();
char *viscsfnk3();
char *viscsfnk4();
char *visam1();
char *visam3();
char *visifosh();
char *viswave();
